// ========================================== //
//                                            //
//       RC6   //
//             2017-12-21               //
//                                            //
//                HiAsm                 //
//              www.hiasm.com                 //
//                                            //
//             : Netspirit               //
//         E-mail: netspirit@meta.ua          //
//                                            //
// ========================================== //

// ****************************************** //
//                RC6 Algorithm               //
//              Copyright (C) 1998            //
//     R. L. Rivest, M. Robshaw, R. Sidney    //
//                RSA Security                //
// ****************************************** //


unit hiRC6Cipher;



interface





uses 
  Kol, BlockCipher;


const
  ROUNDS       = 20; //  .  ,   ,   
  
  BLOCK_SIZE   = 16;
  WORD_SIZE    = 4;
  SUBKEY_COUNT = 2 * (ROUNDS + 2);

type
  
  TRC6Block = array[0..(BLOCK_SIZE div WORD_SIZE)-1] of LongWord;
  
  TWordsArray = array of LongWord;
  
  ThiRC6Cipher = class(ThiBlockCipher)
    private
      FSubKeys: array[0..SUBKEY_COUNT-1] of LongWord;
      
    public
      constructor Create;
      procedure InitializeKey(Key: Pointer; KeyLen: Integer); override;
      procedure EncryptBlock(var Block); override;
      procedure DecryptBlock(var Block); override;
  end;





implementation



const
  P32       = $b7e15163;
  Q32       = $9e3779b9;
  lgw       = 5;



{$ifdef FPC}
procedure ROLBlock(var B: TRC6Block);
var
  t: LongWord;
begin
  t := B[0];
  B[0] := B[1];
  B[1] := B[2];
  B[2] := B[3];
  B[3] := t;
end;

procedure RORBlock(var B: TRC6Block);
var
  t: LongWord;
begin
  t := B[3];
  B[3] := B[2];
  B[2] := B[1];
  B[1] := B[0];
  B[0] := t;
end;
{$endif}



constructor ThiRC6Cipher.Create;
begin
  FBlockSize := BLOCK_SIZE;
  inherited;
end;

procedure ThiRC6Cipher.InitializeKey(Key: Pointer; KeyLen: Integer);
var
  i, j, k, c: Integer;
  A, B: LongWord;
  L: TWordsArray;
begin

  //       'c' 
  if KeyLen > 0 then
  begin
    c := (KeyLen div WORD_SIZE);
    if (c * WORD_SIZE) < KeyLen then Inc(c); //       -    
  end
  else
    c := 1;
  
  SetLength(L, c);
  L[c - 1] := 0; //    (       )
  
  Move(Key^, L[0], KeyLen);
  
  
  //   

  //  
  FSubKeys[0] := P32;
  for i := 1 to SUBKEY_COUNT-1 do
    FSubKeys[i] := FSubKeys[i-1] + Q32;  
  
  //    
  
  i := 0;
  j := 0;
  A := 0;
  B := 0;

  for k := 1 to 3*Max(c, SUBKEY_COUNT) do
  begin
    A := ROLi((FSubKeys[i] + A + B), 3);
    FSubKeys[i] := A;
    B := ROL((L[j] + A + B), (A + B));
    L[j] := B;
    i := (i + 1) mod SUBKEY_COUNT;
    j := (j + 1) mod c;
  end;

end;

procedure ThiRC6Cipher.EncryptBlock(var Block);
var
  B: TRC6Block absolute Block;
  i: Integer;
  t, u: LongWord;
begin
  Inc(B[1], FSubKeys[0]);
  Inc(B[3], FSubKeys[1]);

  for i := 1 to ROUNDS do
  begin
    t := ROLi((B[1] * (2*B[1] + 1)), lgw);
    u := ROLi((B[3] * (2*B[3] + 1)), lgw);
    B[0] := ROL((B[0] xor t), u) + FSubKeys[2*i];
    B[2] := ROL((B[2] xor u), t) + FSubKeys[2*i+1];

    {$ifdef FPC}
    ROLBlock(B);
    {$else}
    t := B[0];
    B[0] := B[1];
    B[1] := B[2];
    B[2] := B[3];
    B[3] := t;
    {$endif}
  end;

  Inc(B[0], FSubKeys[2*ROUNDS+2]);
  Inc(B[2], FSubKeys[2*ROUNDS+3]);
end;

procedure ThiRC6Cipher.DecryptBlock(var Block);
var
  B: TRC6Block absolute Block;
  i: Integer;
  t, u: LongWord;
begin
  Dec(B[2], FSubKeys[2*ROUNDS+3]);
  Dec(B[0], FSubKeys[2*ROUNDS+2]);

  for i := ROUNDS downto 1 do
  begin
    {$ifdef FPC}
    RORBlock(B);
    {$else}
    t := B[3];
    B[3] := B[2];
    B[2] := B[1];
    B[1] := B[0];
    B[0] := t;
    {$endif}

    u := ROLi((B[3] * (2*B[3] + 1)), lgw);
    t := ROLi((B[1] * (2*B[1] + 1)), lgw);
    B[2] := ROR((B[2] - FSubKeys[2*i+1]), t) xor u;
    B[0] := ROR((B[0] - FSubKeys[2*i]), u) xor t;
  end;

  Dec(B[3], FSubKeys[1]);
  Dec(B[1], FSubKeys[0]);
end;


end.